@@ -20,18 +20,19 @@ import gzip
 import os
 import pathlib
 import pickle
-import sys
 
 import pytest
 import weakref
 
 import pyarrow as pa
 from pyarrow.tests.test_io import assert_file_not_found
-from pyarrow.vendored.version import Version
+from pyarrow.tests.util import (_filesystem_uri, ProxyHandler,
+                                _configure_s3_limited_user)
 
 from pyarrow.fs import (FileType, FileInfo, FileSelector, FileSystem,
                         LocalFileSystem, SubTreeFileSystem, _MockFileSystem,
-                        FileSystemHandler, PyFileSystem, FSSpecHandler)
+                        FileSystemHandler, PyFileSystem, FSSpecHandler,
+                        copy_files)
 
 
 class DummyHandler(FileSystemHandler):
@@ -97,7 +98,7 @@ class DummyHandler(FileSystemHandler):
     def delete_dir(self, path):
         assert path == "delete_dir"
 
-    def delete_dir_contents(self, path):
+    def delete_dir_contents(self, path, missing_dir_ok):
         if not path.strip("/"):
             raise ValueError
         assert path == "delete_dir_contents"
@@ -139,67 +140,6 @@ class DummyHandler(FileSystemHandler):
         return pa.BufferOutputStream()
 
 
-class ProxyHandler(FileSystemHandler):
-
-    def __init__(self, fs):
-        self._fs = fs
-
-    def __eq__(self, other):
-        if isinstance(other, ProxyHandler):
-            return self._fs == other._fs
-        return NotImplemented
-
-    def __ne__(self, other):
-        if isinstance(other, ProxyHandler):
-            return self._fs != other._fs
-        return NotImplemented
-
-    def get_type_name(self):
-        return "proxy::" + self._fs.type_name
-
-    def normalize_path(self, path):
-        return self._fs.normalize_path(path)
-
-    def get_file_info(self, paths):
-        return self._fs.get_file_info(paths)
-
-    def get_file_info_selector(self, selector):
-        return self._fs.get_file_info(selector)
-
-    def create_dir(self, path, recursive):
-        return self._fs.create_dir(path, recursive=recursive)
-
-    def delete_dir(self, path):
-        return self._fs.delete_dir(path)
-
-    def delete_dir_contents(self, path):
-        return self._fs.delete_dir_contents(path)
-
-    def delete_root_dir_contents(self):
-        return self._fs.delete_dir_contents("", accept_root_dir=True)
-
-    def delete_file(self, path):
-        return self._fs.delete_file(path)
-
-    def move(self, src, dest):
-        return self._fs.move(src, dest)
-
-    def copy_file(self, src, dest):
-        return self._fs.copy_file(src, dest)
-
-    def open_input_stream(self, path):
-        return self._fs.open_input_stream(path)
-
-    def open_input_file(self, path):
-        return self._fs.open_input_file(path)
-
-    def open_output_stream(self, path, metadata):
-        return self._fs.open_output_stream(path, metadata=metadata)
-
-    def open_append_stream(self, path, metadata):
-        return self._fs.open_append_stream(path, metadata=metadata)
-
-
 @pytest.fixture
 def localfs(request, tempdir):
     return dict(
@@ -261,11 +201,11 @@ def subtree_localfs(request, tempdir, localfs):
 
 
 @pytest.fixture
-def s3fs(request, s3_connection, s3_server):
+def s3fs(request, s3_server):
     request.config.pyarrow.requires('s3')
     from pyarrow.fs import S3FileSystem
 
-    host, port, access_key, secret_key = s3_connection
+    host, port, access_key, secret_key = s3_server['connection']
     bucket = 'pyarrow-filesystem/'
 
     fs = S3FileSystem(
@@ -296,6 +236,28 @@ def subtree_s3fs(request, s3fs):
     )
 
 
+_minio_limited_policy = """{
+    "Version": "2012-10-17",
+    "Statement": [
+        {
+            "Effect": "Allow",
+            "Action": [
+                "s3:ListAllMyBuckets",
+                "s3:PutObject",
+                "s3:GetObject",
+                "s3:ListBucket",
+                "s3:PutObjectTagging",
+                "s3:DeleteObject",
+                "s3:GetObjectVersion"
+            ],
+            "Resource": [
+                "arn:aws:s3:::*"
+            ]
+        }
+    ]
+}"""
+
+
 @pytest.fixture
 def hdfs(request, hdfs_connection):
     request.config.pyarrow.requires('hdfs')
@@ -343,13 +305,9 @@ def py_fsspec_memoryfs(request, tempdir):
 
 
 @pytest.fixture
-def py_fsspec_s3fs(request, s3_connection, s3_server):
+def py_fsspec_s3fs(request, s3_server):
     s3fs = pytest.importorskip("s3fs")
-    if (sys.version_info < (3, 7) and
-            Version(s3fs.__version__) >= Version("0.5")):
-        pytest.skip("s3fs>=0.5 version is async and requires Python >= 3.7")
-
-    host, port, access_key, secret_key = s3_connection
+    host, port, access_key, secret_key = s3_server['connection']
     bucket = 'pyarrow-filesystem/'
 
     fs = s3fs.S3FileSystem(
@@ -384,11 +342,13 @@ def py_fsspec_s3fs(request, s3_connection, s3_server):
     ),
     pytest.param(
         pytest.lazy_fixture('s3fs'),
-        id='S3FileSystem'
+        id='S3FileSystem',
+        marks=pytest.mark.s3
     ),
     pytest.param(
         pytest.lazy_fixture('hdfs'),
-        id='HadoopFileSystem'
+        id='HadoopFileSystem',
+        marks=pytest.mark.hdfs
     ),
     pytest.param(
         pytest.lazy_fixture('mockfs'),
@@ -412,7 +372,8 @@ def py_fsspec_s3fs(request, s3_connection, s3_server):
     ),
     pytest.param(
         pytest.lazy_fixture('py_fsspec_s3fs'),
-        id='PyFileSystem(FSSpecHandler(s3fs.S3FileSystem()))'
+        id='PyFileSystem(FSSpecHandler(s3fs.S3FileSystem()))',
+        marks=pytest.mark.s3
     ),
 ])
 def filesystem_config(request):
@@ -468,6 +429,21 @@ def skip_fsspec_s3fs(fs):
         pytest.xfail(reason="Not working with fsspec's s3fs")
 
 
+@pytest.mark.s3
+def test_s3fs_limited_permissions_create_bucket(s3_server):
+    from pyarrow.fs import S3FileSystem
+    _configure_s3_limited_user(s3_server, _minio_limited_policy)
+    host, port, _, _ = s3_server['connection']
+
+    fs = S3FileSystem(
+        access_key='limited',
+        secret_key='limited123',
+        endpoint_override='{}:{}'.format(host, port),
+        scheme='http'
+    )
+    fs.create_dir('existing-bucket/test')
+
+
 def test_file_info_constructor():
     dt = datetime.fromtimestamp(1568799826, timezone.utc)
 
@@ -528,10 +504,14 @@ def test_subtree_filesystem():
     subfs = SubTreeFileSystem('/base', localfs)
     assert subfs.base_path == '/base/'
     assert subfs.base_fs == localfs
+    assert repr(subfs).startswith('SubTreeFileSystem(base_path=/base/, '
+                                  'base_fs=<pyarrow._fs.LocalFileSystem')
 
     subfs = SubTreeFileSystem('/another/base/', LocalFileSystem())
     assert subfs.base_path == '/another/base/'
     assert subfs.base_fs == localfs
+    assert repr(subfs).startswith('SubTreeFileSystem(base_path=/another/base/,'
+                                  ' base_fs=<pyarrow._fs.LocalFileSystem')
 
 
 def test_filesystem_pickling(fs):
@@ -755,6 +735,9 @@ def test_delete_dir_contents(fs, pathfn):
     fs.delete_dir_contents(d)
     with pytest.raises(pa.ArrowIOError):
         fs.delete_dir(nd)
+    fs.delete_dir_contents(nd, missing_dir_ok=True)
+    with pytest.raises(pa.ArrowIOError):
+        fs.delete_dir_contents(nd)
     fs.delete_dir(d)
     with pytest.raises(pa.ArrowIOError):
         fs.delete_dir(d)
@@ -893,6 +876,13 @@ def test_open_input_file(fs, pathfn):
     assert result == data[read_from:]
 
 
+def test_open_input_stream_not_found(fs, pathfn):
+    # The proper exception should be raised for this common case (ARROW-15896)
+    p = pathfn('open-input-stream-not-found')
+    with pytest.raises(FileNotFoundError):
+        fs.open_input_stream(p)
+
+
 @pytest.mark.gzip
 @pytest.mark.parametrize(
     ('compression', 'buffer_size', 'decompressor'),
@@ -1313,10 +1303,10 @@ def test_filesystem_from_path_object(path):
 
 
 @pytest.mark.s3
-def test_filesystem_from_uri_s3(s3_connection, s3_server):
+def test_filesystem_from_uri_s3(s3_server):
     from pyarrow.fs import S3FileSystem
 
-    host, port, access_key, secret_key = s3_connection
+    host, port, access_key, secret_key = s3_server['connection']
 
     uri = "s3://{}:{}@mybucket/foo/bar?scheme=http&endpoint_override={}:{}" \
         .format(access_key, secret_key, host, port)
@@ -1542,3 +1532,120 @@ def test_s3_real_aws_region_selection():
     fs, path = FileSystem.from_uri(
         's3://x-arrow-non-existent-bucket?region=us-east-3')
     assert fs.region == 'us-east-3'
+
+
+@pytest.mark.s3
+def test_resolve_s3_region():
+    from pyarrow.fs import resolve_s3_region
+    assert resolve_s3_region('ursa-labs-taxi-data') == 'us-east-2'
+    assert resolve_s3_region('mf-nwp-models') == 'eu-west-1'
+
+    with pytest.raises(ValueError, match="Not a valid bucket name"):
+        resolve_s3_region('foo/bar')
+    with pytest.raises(ValueError, match="Not a valid bucket name"):
+        resolve_s3_region('s3:bucket')
+
+
+@pytest.mark.s3
+def test_copy_files(s3_connection, s3fs, tempdir):
+    fs = s3fs["fs"]
+    pathfn = s3fs["pathfn"]
+
+    # create test file on S3 filesystem
+    path = pathfn('c.txt')
+    with fs.open_output_stream(path) as f:
+        f.write(b'test')
+
+    # create URI for created file
+    host, port, access_key, secret_key = s3_connection
+    source_uri = (
+        f"s3://{access_key}:{secret_key}@{path}"
+        f"?scheme=http&endpoint_override={host}:{port}"
+    )
+    # copy from S3 URI to local file
+    local_path1 = str(tempdir / "c_copied1.txt")
+    copy_files(source_uri, local_path1)
+
+    localfs = LocalFileSystem()
+    with localfs.open_input_stream(local_path1) as f:
+        assert f.read() == b"test"
+
+    # copy from S3 path+filesystem to local file
+    local_path2 = str(tempdir / "c_copied2.txt")
+    copy_files(path, local_path2, source_filesystem=fs)
+    with localfs.open_input_stream(local_path2) as f:
+        assert f.read() == b"test"
+
+    # copy to local file with URI
+    local_path3 = str(tempdir / "c_copied3.txt")
+    destination_uri = _filesystem_uri(local_path3)  # file://
+    copy_files(source_uri, destination_uri)
+
+    with localfs.open_input_stream(local_path3) as f:
+        assert f.read() == b"test"
+
+    # copy to local file with path+filesystem
+    local_path4 = str(tempdir / "c_copied4.txt")
+    copy_files(source_uri, local_path4, destination_filesystem=localfs)
+
+    with localfs.open_input_stream(local_path4) as f:
+        assert f.read() == b"test"
+
+    # copy with additional options
+    local_path5 = str(tempdir / "c_copied5.txt")
+    copy_files(source_uri, local_path5, chunk_size=1, use_threads=False)
+
+    with localfs.open_input_stream(local_path5) as f:
+        assert f.read() == b"test"
+
+
+def test_copy_files_directory(tempdir):
+    localfs = LocalFileSystem()
+
+    # create source directory with 2 files
+    source_dir = tempdir / "source"
+    source_dir.mkdir()
+    with localfs.open_output_stream(str(source_dir / "file1")) as f:
+        f.write(b'test1')
+    with localfs.open_output_stream(str(source_dir / "file2")) as f:
+        f.write(b'test2')
+
+    def check_copied_files(destination_dir):
+        with localfs.open_input_stream(str(destination_dir / "file1")) as f:
+            assert f.read() == b"test1"
+        with localfs.open_input_stream(str(destination_dir / "file2")) as f:
+            assert f.read() == b"test2"
+
+    # Copy directory with local file paths
+    destination_dir1 = tempdir / "destination1"
+    # TODO need to create?
+    destination_dir1.mkdir()
+    copy_files(str(source_dir), str(destination_dir1))
+    check_copied_files(destination_dir1)
+
+    # Copy directory with path+filesystem
+    destination_dir2 = tempdir / "destination2"
+    destination_dir2.mkdir()
+    copy_files(str(source_dir), str(destination_dir2),
+               source_filesystem=localfs, destination_filesystem=localfs)
+    check_copied_files(destination_dir2)
+
+    # Copy directory with URI
+    destination_dir3 = tempdir / "destination3"
+    destination_dir3.mkdir()
+    source_uri = _filesystem_uri(str(source_dir))  # file://
+    destination_uri = _filesystem_uri(str(destination_dir3))
+    copy_files(source_uri, destination_uri)
+    check_copied_files(destination_dir3)
+
+    # Copy directory with Path objects
+    destination_dir4 = tempdir / "destination4"
+    destination_dir4.mkdir()
+    copy_files(source_dir, destination_dir4)
+    check_copied_files(destination_dir4)
+
+    # copy with additional non-default options
+    destination_dir5 = tempdir / "destination5"
+    destination_dir5.mkdir()
+    copy_files(source_dir, destination_dir5, chunk_size=1, use_threads=False)
+    check_copied_files(destination_dir5)
