@@ -327,6 +327,16 @@ def test_python_file_closing():
 # Buffers
 
 
+def check_buffer_pickling(buf):
+    # Check that buffer survives a pickle roundtrip
+    for protocol in range(0, pickle.HIGHEST_PROTOCOL + 1):
+        result = pickle.loads(pickle.dumps(buf, protocol=protocol))
+        assert len(result) == len(buf)
+        assert memoryview(result) == memoryview(buf)
+        assert result.to_pybytes() == buf.to_pybytes()
+        assert result.is_mutable == buf.is_mutable
+
+
 def test_buffer_bytes():
     val = b'some data'
 
@@ -336,13 +346,22 @@ def test_buffer_bytes():
     assert buf.is_cpu
 
     result = buf.to_pybytes()
-
     assert result == val
 
-    # Check that buffers survive a pickle roundtrip
-    result_buf = pickle.loads(pickle.dumps(buf))
-    result = result_buf.to_pybytes()
-    assert result == val
+    check_buffer_pickling(buf)
+
+
+def test_buffer_null_data():
+    null_buff = pa.foreign_buffer(address=0, size=0)
+    assert null_buff.to_pybytes() == b""
+    assert null_buff.address == 0
+    # ARROW-16048: we shouldn't expose a NULL address through the Python
+    # buffer protocol.
+    m = memoryview(null_buff)
+    assert m.tobytes() == b""
+    assert pa.py_buffer(m).address != 0
+
+    check_buffer_pickling(null_buff)
 
 
 def test_buffer_memoryview():
@@ -354,9 +373,10 @@ def test_buffer_memoryview():
     assert buf.is_cpu
 
     result = memoryview(buf)
-
     assert result == val
 
+    check_buffer_pickling(buf)
+
 
 def test_buffer_bytearray():
     val = bytearray(b'some data')
@@ -367,9 +387,10 @@ def test_buffer_bytearray():
     assert buf.is_cpu
 
     result = bytearray(buf)
-
     assert result == val
 
+    check_buffer_pickling(buf)
+
 
 def test_buffer_invalid():
     with pytest.raises(TypeError,
@@ -518,10 +539,25 @@ def test_buffer_slicing():
     with pytest.raises(IndexError):
         buf.slice(-1)
 
+    with pytest.raises(IndexError):
+        buf.slice(len(buf) + 1)
+    assert buf[11:].to_pybytes() == b""
+
+    # Slice stop exceeds buffer length
+    with pytest.raises(IndexError):
+        buf.slice(1, len(buf))
+    assert buf[1:11].to_pybytes() == buf.to_pybytes()[1:]
+
+    # Negative length
+    with pytest.raises(IndexError):
+        buf.slice(1, -1)
+
     # Test slice notation
     assert buf[2:].equals(buf.slice(2))
     assert buf[2:5].equals(buf.slice(2, 3))
     assert buf[-5:].equals(buf.slice(len(buf) - 5))
+    assert buf[-5:-2].equals(buf.slice(len(buf) - 5, 3))
+
     with pytest.raises(IndexError):
         buf[::-1]
     with pytest.raises(IndexError):
@@ -639,7 +675,7 @@ def test_compression_level(compression):
         pytest.skip("{} support is not built".format(compression))
 
     # These codecs do not support a compression level
-    no_level = ['snappy', 'lz4']
+    no_level = ['snappy']
     if compression in no_level:
         assert not Codec.supports_compression_level(compression)
         with pytest.raises(ValueError):
@@ -686,7 +722,7 @@ def test_compression_level(compression):
     # The ability to set a seed this way is not present on older versions of
     # numpy (currently in our python 3.6 CI build).  Some inputs might just
     # happen to compress the same between the two levels so using seeded
-    # random numbers is neccesary to help get more reliable results
+    # random numbers is necessary to help get more reliable results
     #
     # The goal of this part is to ensure the compression_level is being
     # passed down to the C++ layer, not to verify the compression algs
@@ -1286,6 +1322,33 @@ def test_compressed_input_bz2(tmpdir):
         pytest.skip(str(e))
 
 
+@pytest.mark.gzip
+def test_compressed_input_openfile(tmpdir):
+    if not Codec.is_available("gzip"):
+        pytest.skip("gzip support is not built")
+
+    data = b"some test data\n" * 10 + b"eof\n"
+    fn = str(tmpdir / "test_compressed_input_openfile.gz")
+    with gzip.open(fn, "wb") as f:
+        f.write(data)
+
+    with pa.CompressedInputStream(fn, "gzip") as compressed:
+        buf = compressed.read_buffer()
+        assert buf.to_pybytes() == data
+    assert compressed.closed
+
+    with pa.CompressedInputStream(pathlib.Path(fn), "gzip") as compressed:
+        buf = compressed.read_buffer()
+        assert buf.to_pybytes() == data
+    assert compressed.closed
+
+    f = open(fn, "rb")
+    with pa.CompressedInputStream(f, "gzip") as compressed:
+        buf = compressed.read_buffer()
+        assert buf.to_pybytes() == data
+    assert f.closed
+
+
 def check_compressed_concatenated(data, fn, compression):
     raw = pa.OSFile(fn, mode="rb")
     with pa.CompressedInputStream(raw, compression) as compressed:
